#!/usr/bin/env python
import random
import numpy as np
import itertools
import matplotlib.pyplot as plt
from scipy.optimize import minimize
from functools import partial
from scipy.optimize import differential_evolution as devo


def evaluate_acquisition_function(x, gp, acquisition_function,origin = None,
        cost_function = None, cost_function_parameters = None):
    ##########################################################
    ####this function evaluates a default or a user-defined acquisition function
    ##########################################################
    if x.ndim == 1:x = np.array([x])
    if cost_function is not None and origin is not None and cost_function_parameters is not None:
        cost_eval = cost_function(origin,x[0],cost_function_parameters)
    else:
        cost_eval = 1.0
    #for user defined acquisition function
    if callable(acquisition_function):
        return -acquisition_function(x,gp)/cost_eval
    obj_eval = evaluate_gp_acquisition_function(x, acquisition_function, gp)
    #if no user defined acquisition function is used
    obj_eval = obj_eval / cost_eval
    return -obj_eval

def evaluate_acquisition_function_gradient(x, gp, acquisition_function, origin = None,
        cost_function = None, cost_function_parameters = None):
    acquisition_gradient = gradient(evaluate_acquisition_function, x, 1e-5, gp,acquisition_function,origin,cost_function,cost_function_parameters)
    return acquisition_gradient

def evaluate_acquisition_function_hessian(x, gp, acquisition_function,origin = None,
        cost_function = None, cost_function_parameters = None):
    acquisition_hessian = hessian(evaluate_acquisition_function, x, 1e-5, gp,acquisition_function,origin,cost_function,cost_function_parameters)
    return acquisition_hessian

def evaluate_gp_acquisition_function(x,acquisition_function,gp):
    ##this function will always spit out a 1d numpy array
    ##for certain functions, this array will only have one entry
    ##for the other the length == len(x)
    if len(x.shape) == 1: x = np.array([x])
    if acquisition_function == "variance":
        #x = cast_to_index_set(x,gp.value_positions[-1], mode = 'cartesian product')
        res = gp.posterior_covariance(x)
        return res
    if acquisition_function == "covariance":
        #x = cast_to_index_set(x,gp.value_positions[-1], mode = 'cartesian product')
        res = gp.posterior_covariance(x)
        b = res["S(x)"]
        sgn, logdet = np.linalg.slogdet(b)
        return np.array([np.sqrt(sgn * np.exp(logdet))])
    ###################more here: shannon_ig  for instance
    elif acquisition_function == "shannon_ig":
        #x = cast_to_index_set(x,gp.value_positions[-1], mode = 'cartesian product')
        res = gp.shannon_information_gain(x)["sig"]
        return np.array([res])
    elif acquisition_function == "upper_confidence":
        #x = cast_to_index_set(x,gp.value_positions[-1], mode = 'cartesian product')
        m = gp.posterior_mean(x)["f(x)"]
        v = gp.posterior_covariance(x)["v(x)"]
        return m + 3.0*v
    elif acquisition_function == "maximum":
        #x = cast_to_index_set(x,gp.value_positions[-1], mode = 'cartesian product')
        res = gp.posterior_mean(x)["f(x)"]
        return res
    elif acquisition_function == "minimum":
        #x = cast_to_index_set(x,gp.value_positions[-1], mode = 'cartesian product')
        res = gp.posterior_mean(x)["f(x)"]
        return -res

##########################################################################
def find_acquisition_function_maxima(gp,acquisition_function,
        origin,number_of_maxima_sought,
        optimization_bounds,
        optimization_method = "global",
        optimization_pop_size = 20,
        optimization_max_iter = 200,
        optimization_tol = 10e-6,
        optimization_x0 = None,
        cost_function = None,
        cost_function_parameters = None,
        dask_client = False):
    bounds = np.array(optimization_bounds)
    print("====================================")
    print("finding acquisition function maxima...")
    print("optimization method ",optimization_method)
    print("tolerance: ", optimization_tol)
    print("population size: ", optimization_pop_size)
    print("maximum number of iterations: ",optimization_max_iter)
    print("bounds: ")
    print(bounds)
    print("cost function parameters: ", cost_function_parameters)
    print("====================================")

    if optimization_method == "global":
        opti, func_eval = differential_evolution(
            evaluate_acquisition_function,
            optimization_bounds,
            tol = optimization_tol,
            popsize = optimization_pop_size,
            max_iter = optimization_max_iter,
            origin = origin,
            gp = gp,
            acquisition_function = acquisition_function,
            cost_function = cost_function,
            cost_function_parameters = cost_function_parameters
        )
        opti = np.asarray(opti)
        func_eval = np.asarray(func_eval)
    elif optimization_method == "hgdl":
        ###run differential evo first if hxdy only returns stationary points
        ###then of hgdl is successful, stack results and return
        from hgdl.hgdl import HGDL
        a = HGDL(evaluate_acquisition_function,
                    evaluate_acquisition_function_gradient,
                    evaluate_acquisition_function_hessian,
                    optimization_bounds, number_of_walkers = optimization_pop_size,
                    verbose = False, maxEpochs = optimization_max_iter,
                    x0 = optimization_x0,
                    args = (gp,acquisition_function,origin,cost_function,cost_function_parameters))
        #####optimization_max_iter, tolerance here
        a.optimize(dask_client = dask_client)
        res = a.get_latest(number_of_maxima_sought)
        opti = res['x']
        func_eval = res['func evals']
    elif optimization_method == "local":
        if optimization_x0 is not None and optimization_x0.ndim == 1: x0 = optimization_x0
        elif optimization_x0 is not None and optimization_x0.ndim == 2: x0 = optimization_x0[0]
        else: x0 = np.random.uniform(low = bounds[:,0],high = bounds[:,1],size = len(bounds))
        a = minimize(
            evaluate_acquisition_function,
            x0,
            args = (gp,acquisition_function,origin,cost_function,cost_function_parameters),
            method="L-BFGS-B",
            jac=evaluate_acquisition_function_gradient,
            bounds = bounds,
            tol = optimization_tol,
            callback = None,
            options = {"maxiter": optimization_max_iter}
            )
        opti = np.array([a["x"]])
        func_eval = np.array(a["fun"])
        if a["success"] is False:
            print("local acquisition function optimization not successful, solution replaced with random point.")
            opti = np.array(x0)
            if opti.ndim !=2 : opti = np.array([opti])
            func_eval = evaluate_acquisition_function(x0,
                    gp,acquisition_function,origin,
                    cost_function,cost_function_parameters)
            if func_eval.ndim != 1: func_eval = np.array([func_eval])
    else:
        raise ValueError("Invalid acquisition function optimization method given.")

    if func_eval.ndim != 1 or opti.ndim != 2:
        print("f(x): ",func_eval)
        print("x: ",opti)
        raise Exception("The output of the acquisition function optimization is not 2 dimensional. Please check your acquisition function.")
    return opti,func_eval

############################################################
############################################################
############################################################
############################################################
def differential_evolution(ObjectiveFunction, bounds, tol, popsize, max_iter = 100, 
        origin = None, gp = None, acquisition_function = None, cost_function = None, cost_function_parameters = None):
    fun =  partial(ObjectiveFunction, gp = gp, acquisition_function = acquisition_function, origin = origin,
            cost_function = cost_function, cost_function_parameters = cost_function_parameters)
    res = devo(
        fun, bounds, tol=tol, disp=True, maxiter=max_iter, popsize=popsize, polish=False
    )
    return [list(res["x"])], list([res["fun"]])


def normed_gaussian_function(x, mean, sigma2):
    return (1.0 / np.sqrt(2.0 * np.pi * sigma2)) * np.exp(
        -((x - mean) ** 2) / (2.0 * sigma2)
    )
def gradient(function, point, epsilon = 1e-6,*args):
    """
    This function calculates the gradient of a function by using finite differences

    Extended description of function.

    Parameters:
    function (function object): the function the gradient should be computed of
    point (numpy array 1d): point at which the gradient should be computed

    optional:
    epsilon (float): the distance used for the evaluation of the function

    Returns:
    numpy array of gradient

    """
    gradient = np.zeros((len(point)))
    #args = args[0]
    for i in range(len(point)):
        new_point = np.array(point)
        new_point[i] += epsilon
        gradient[i] = (function(new_point,*args) - function(point,*args))/ epsilon
    return gradient

def hessian(function, point, epsilon = 1e-3, *args):
    """
    This function calculates the hessian of a function by using finite differences

    Extended description of function.

    Parameters:
    function (function object): the function, the hessian should be computed of
    point (numpy array 1d): point at which the gradient should be computed

    optional:
    epsilon (float): the distance used for the evaluation of the function

    Returns:
    numpy array of hessian

    """
    hessian = np.zeros((len(point),len(point)))
    for i in range(len(point)):
        for j in range(len(point)):
            new_point1 = np.array(point)
            new_point2 = np.array(point)
            new_point3 = np.array(point)
            new_point4 = np.array(point)

            new_point1[i] = new_point1[i] + epsilon
            new_point1[j] = new_point1[j] + epsilon

            new_point2[i] = new_point2[i] + epsilon
            new_point2[j] = new_point2[j] - epsilon

            new_point3[i] = new_point3[i] - epsilon
            new_point3[j] = new_point3[j] + epsilon

            new_point4[i] = new_point4[i] - epsilon
            new_point4[j] = new_point4[j] - epsilon

            hessian[i,j] = \
            (function(new_point1,*args) - function(new_point2,*args) - function(new_point3,*args) +  function(new_point4,*args))\
            / (4.0*(epsilon**2))
    return hessian


def bhattacharyya_distance(reference_distribution, test_distribution, dx):
    y1 = reference_distribution / (sum(reference_distribution)*dx)
    y2 = test_distribution / (sum(test_distribution)*dx)
    return sum(np.sqrt(y1*y2)) * dx


def cast_to_index_set(x_input,x_output,mode = 'cartesian product'):
    n_orig = len(x_input)
    tasks = len(x_output)
    if mode == 'cartesian product':
        new_points = np.zeros((len(x_input) * len(x_output), len(x_input[0]) + len(x_output[0])))
        counter = 0
        for element in itertools.product(x_input, x_output):
            new_points[counter] = np.concatenate([element[0], element[1]], axis=0)
            counter += 1   ###can't we append?
    elif mode == 'stack':
        new_points = np.column_stack([x_input,x_output])
    return new_points




import numpy as np
from fvgp.fvgp import FVGP
from . import surrogate_model as sm
import time


class GPExpertsOptimizer():
    """
    This class can combine a set of Gaussian Experts (GPOptimizer objects)
    and combine them.
    This es especially usefull for large-scale GPs
    
    Attributes:
        a list of GPOptimizer Objects
    """
    def __init__(self,gp_optimizer_obj_list):
        self.gpol = gp_optimizer_obj_list
        self.kl_div = self.gpol[0].gp.kl_div
        self.kl_div_grad = self.gpol[0].gp.kl_div_grad
        self.entropy = self.gpol[0].gp.entropy

    def predict(x):
        S = self.gpol[0].gp.posterior_covariance["S(x)"]
        mu = self.gpol[0].gp.posterior_mean["f(x)"]
        for i in range(1,len(self.gpol)):
            cov = self.gpol[i].gp.posterior_covariance["S(x)"]
            inv = np.linalg.inv(S + cov)
            mu = cov @ inv @ mu + S @ inv @ self.gpol[i].gp.posterior_mean["f(x)"]
            S  = S @ inv @ cov
        return {"f(x)":mu, "S(x)": S}

    def ask(self, position = None, n = 1,
            acquisition_function = "covariance",
            optimization_bounds = None,
            optimization_method = "global",
            optimization_pop_size = 20,
            optimization_max_iter = 20,
            optimization_tol = 10e-6,
            dask_client = False):
    return 0




#!/usr/bin/env python

import numpy as np
from fvgp.fvgp import fvGP
from gpcam import surrogate_model as sm
from fvgp.gp import GP

class GPOptimizer(GP):
    """
    GPOptimizer class: Given data, this class can determine which
    data should be collected next.
    Initialize and then use tell() to communicate data.
    Use init_gp() to initlaize a GP.
    Use ask() to ask for the optimal next point

    symbols:
        N: Number of points in the data set
        n: number of return values
        dim1: number of dimension of the input space
        dim2: number of dimension of the output space

    Attributes:
        input_space_dimension (int):         dim1
        input_space_bounds (2d array):         bounds of the index set


    Example:
        obj = GPOptimizer(3,[[0,10],[0,10],[0,10]])
        obj.tell(x,y)
        obj.init_gp(...)
        obj.train_gp(...) #can be "train_gp_async()"
        co = obj.init_cost(...)
        ask(cost_function = co)
        obj.update_cost_function(...)
        prediction = obj.posterior_mean(x0)
    ------------------------------------------------------------
    """

    def __init__(
        self,
        input_space_dimension,
        input_space_bounds,
        ):
        """
        GPOptimizer constructor
        type help(gp_optimizer) for help
        """
        self.iput_dim = input_space_dimension
        self.points = np.empty((0, self.iput_dim))
        self.values = np.empty((0))
        self.variances = np.empty((0))
        self.input_space_bounds = np.array(input_space_bounds)
        self.gp_initialized = False
        self.cost_function_parameters = None
        self.cost_function = None
        self.consider_costs = False

    def get_data(self):
        """
        Provides a way to access the current class attributes.

        Returns
        -------
        dict
            Dictionary containing the input dim, output dim, output number,
            x & y data, measurement variances, measurement value positions,
            hyperparameters, cost function parameters and consider costs
            class attributes. Note that if tell() has not been called, many
            of these returned values will be None.
        """
        if self.gp_initialized: hps = self.hyperparameters
        else: hps = None
        return {
            "input dim": self.iput_dim,
            "x": self.points,
            "y": self.values,
            "measurement variances": self.variances,
            "hyperparameters": hps,
            "cost function parameters": self.cost_function_parameters,
            "consider costs": self.consider_costs,
            }

    def evaluate_acquisition_function(self,
        x, acquisition_function="covariance", cost_function=None,
        origin=None):
        """
        Evaluates the acquisition function.

        Parameters:
        -----------
        x: 1d numpy array.

        Optional Parameters:
        --------------------
        acquisition_function : default = "covariance",
                               "covariance","shannon_ig" ,..., or callable, use the same you use
                               in ask(). (The default is "covariance").
        origin:                default = None, only important for cost considerations, mandatory if costs are important

        Returns
        -------
        float or numpy array
        """

        if self.gp_initialized is False:
            raise Exception(
                "Initialize GP before evaluating the acquisition function. "
                "See help(gp_init)."
            )
        x = np.array(x)
        try:
            return sm.evaluate_acquisition_function(
                x, self, acquisition_function, origin, cost_function,
                self.cost_function_parameters)
        except Exception as a:
            print("Evaluating the acquisition function was not successful.")
            print("Error Message:")
            print(str(a))

    def tell(self, x, y, variances=None):
        """
        This function can tell() the gp_optimizer class
        the data that was collected. The data will instantly be use to update the gp_data
        if a GP was previously initialized

        Parameters:
        -----------
            x (2d numpy array):                A 2d array of all the points in the data
            values (2d numpy array):           A 2d array of all the values measured at the associated points

        Optional Parameters:
        --------------------
            variances (2d numpy array):         A 2d array of all the variances of the measured points

        Returns:
        --------
            no returns
        """
        self.points = x
        self.values = y
        self.variances = variances
        print("New data communicated via tell()")

        if self.gp_initialized is True: self.update_gp()

##############################################################
    def init_gp(
        self,
        init_hyperparameters,
        compute_device="cpu",
        gp_kernel_function=None,
        gp_mean_function=None,
        sparse=False
    ):
        """
        Function to initialize the GP if it has not already been initialized
        Parameters:
        -----------
            init_hyperparameters:   1d numpy array containing the initial guesses for the hyperparemeters
        Optional Parameters:
        --------------------
            compute_device, default = cpu, others = "gpu"
            gp_kernel_function, default = fvGP default
            gp_mean_function, default = fvGP default, i.e. average of data
            sparse, default = False
        """
        if self.gp_initialized is False:
            GP.__init__(
            self,
            self.iput_dim,
            self.points,
            self.values,
            init_hyperparameters,
            variances = self.variances,
            compute_device = compute_device,
            gp_kernel_function = gp_kernel_function,
            gp_mean_function = gp_mean_function,
            sparse = sparse,
            normalize_y = False
            )
            self.gp_initialized = True
            print("GP successfully initiated")
        else: print("GP already initialized")

##############################################################
    def update_gp(self):
        """
        This function updates the data in the GP, tell(...) will call this function automatically if
        GP is intialized
        Paramaters:
        -----------
            no input parameters
        """
        self.update_gp_data(
            self.points,
            self.values,
            variances = self.variances)
        print("GP data updated")

##############################################################
    def train_gp_async(self,
            hyperparameter_bounds,
            pop_size = 20,
            tolerance = 1e-6,
            max_iter = 10000,
            dask_client = None):
        """
        Function to start fvGP asynchronous training.
        Parameters:
        -----------
            hyperparameter_bounds:                  2d np.array of bounds for the hyperparameters
        Optional Parameters:
        --------------------
            pop_size:       number of walkers in the optimization, default = 20
            tolerance:      tolerance for termination, default = 1e-6
            max_iter:       maximum number of iterations, default = 10000
            dask_client:    a DASK client, see dask package docs for explanation
        Return:
            Nothing, call update_hyperparameters() for the result
        """
        if self.gp_initialized is False:
            raise Exception("No GP to be trained. Please call init_gp(...) before training.")
        opt_obj = self.train_async(
                hyperparameter_bounds,
                init_hyperparameters = self.hyperparameters,
                pop_size = pop_size,
                tolerance = tolerance,
                max_iter = max_iter,
                dask_client = dask_client
                )
        return opt_obj

##############################################################
    def train_gp(self,hyperparameter_bounds,
            method = "global",pop_size = 20,
            optimization_dict = None,tolerance = 1e-6,
            max_iter = 120):
        """
        Function to perform fvGP training.
        Parameters:
        -----------
            hyperparameter_bounds:                  2d np.array of bounds for the hyperparameters
        Optional Parameters:
        --------------------
            method:            "hgdl"/"global"/"local", default = "global"
            pop_size:          number of walkers in the optimization, default = 20
            optimization_dict: default = None
            tolerance:         tolerance for termination, default = 1e-6
            max_iter:          maximum number of iterations, default = 120
        Return:
            trained hyperparameters (just for info, the GP is already updated)
        """
        if self.gp_initialized is False:
            raise Exception("No GP to be trained. Please call init_gp(...) before training.")
        self.train(
                hyperparameter_bounds,
                init_hyperparameters = self.hyperparameters,
                method = method,
                optimization_dict = optimization_dict,
                pop_size = pop_size,
                tolerance = tolerance,
                max_iter = max_iter
                )
        return self.hyperparameters

##############################################################
    def stop_async_train(self, opt_obj):
        """
        function to stop vfGP async training
        Parameters:
        -----------
            no input parameters
        """
        try: self.stop_training(opt_obj)
        except: pass

    def kill_async_train(self, opt_obj):
        """
        function to stop vfGP async training
        Parameters:
        -----------
            no input parameters
        """
        try: self.kill_training(opt_obj)
        except: pass

##############################################################
    def update_hyperparameters(self, opt_obj):
        hps = GP.update_hyperparameters(self, opt_obj)
        print("GPOptimizer updated the Hyperperameters: ", self.hyperparameters)
        return hps

##############################################################
    def ask(self, position = None, n = 1,
            acquisition_function = "covariance",
            cost_function = None,
            bounds = None,
            method = "global",
            pop_size = 20,
            max_iter = 20,
            tol = 10e-6,
            x0 = None,
            dask_client = False):
        """
        Given that the acquisition device is at "position", the function ask() s for
        "n" new optimal points within certain "bounds" and using the optimization setup:
        "acquisition_function_pop_size", "max_iter" and "tol"
        Parameters:
        -----------

        Optional Parameters:
        --------------------
            position (numpy array):            last measured point, default = None
            n (int):                           how many new measurements are requested, default = 1
            acquisition_function:              default = None, means that the class acquisition function will be used
            cost_function:                     default = None, otherwise cost objective received from init_cost, or callable
            bounds (2d list/None):             default = None
            method:                            default = "global", "global"/"hgdl"
            pop_size (int):                    default = 20
            max_iter (int):                    default = 20
            tol (float):                       default = 10e-6
            x0:                                default = None, starting positions for optimizer
            dask_client:                                    default = False
        """
        print("ask() initiated with hyperparameters:",self.hyperparameters)
        print("optimization method: ", method)
        print("bounds: ",bounds)
        if bounds is None: bounds = self.input_space_bounds
        maxima,func_evals = sm.find_acquisition_function_maxima(
                self,
                acquisition_function,
                position,n, bounds,
                optimization_method = method,
                optimization_pop_size = pop_size,
                optimization_max_iter = max_iter,
                optimization_tol = tol,
                cost_function = cost_function,
                cost_function_parameters = self.cost_function_parameters,
                optimization_x0 = x0,
                dask_client = dask_client)
        return {'x':np.array(maxima), "f(x)" : np.array(func_evals)}

##############################################################
    def init_cost(self,cost_function,cost_function_parameters,cost_update_function = None):
        """
        This function initializes the costs. If used, the acquisition function will be augmented by the costs
        which leads to different suggestions

        Parameters:
        -----------
            cost_function: callable
            cost_function_parameters: arbitrary, are passed to the user defined cost function

        Optional Parameters:
        --------------------
            cost_update_function: a function that updates the cost_function_parameters, default = None
            cost_function_optimization_bounds: optimization bounds for the update, default = None

        Return:
        -------
            cost function that can be injected into ask()
        """

        self.cost_function = cost_function
        self.cost_function_parameters = cost_function_parameters
        self.cost_update_function = cost_update_function
        self.consider_costs = True
        print("Costs successfully initialized")
        return self.cost_function

##############################################################
    def update_cost_function(self,measurement_costs):
        """
        This function updates the parameters for the cost function
        It essentially calls the user-given cost_update_function which
        should return the new parameters how they are used by the user defined
        cost function
        Parameters:
        -----------
            measurement_costs:    an arbitrary structure that describes 
                                  the costs when moving in the parameter space
            cost_update_function: a user-defined function 
                                  def name(measurement_costs,
                                  cost_function_optimization_bounds,cost_function_parameters)
                                  which returns the new parameters
            cost_function_optimization_bounds: see above
        Optional Parameters:
        --------------------
            cost_function_parameters, default = None
        """

        print("Performing cost function update...")
        if self.cost_function_parameters is None: raise Exception("No cost function parameters specified. Please call init_cost() first.")
        self.cost_function_parameters = \
        self.cost_update_function(measurement_costs, self.cost_function_parameters)
        print("cost parameters changed to: ", self.cost_function_parameters)
######################################################################################
######################################################################################
######################################################################################
class fvGPOptimizer(fvGP, GPOptimizer):
    """
    fvGPOptimizer class: Given data, this class can determine which
    data should be collected next.
    Initialize and then use tell() to communicate data.
    Use init_gp() to initlaize a GP.
    Use ask() to ask for the optimal next point

    symbols:
        N: Number of points in the data set
        n: number of return values
        dim1: number of dimension of the input space
        dim2: number of dimension of the output space

    Attributes:
        input_space_dimension (int):         dim1
        output_space_dimension (int):        dim2
        output_number (int):                 n
        input_space_bounds (2d array):         bounds of the index set


    Example:
        obj = fvGPOptimizer(3,1,2,[[0,10],[0,10],[0,10]])
        obj.tell(x,y)
        obj.init_gp(...)
        obj.train_gp(...) #can be "train_gp_async()"
        obj.init_cost(...)
        obj.update_cost_function(...)
        prediction = obj.gp.posterior_mean(x0)
    ------------------------------------------------------------
    """

    def __init__(
        self,
        input_space_dimension,
        output_space_dimension,
        output_number,
        input_space_bounds,
    ):
        """
        GPOptimizer constructor
        type help(gp_optimizer) for help
        """

        self.iput_dim = input_space_dimension
        self.oput_dim = output_space_dimension
        self.output_number = output_number
        self.points = np.empty((0, self.iput_dim))
        self.values = np.empty((0, self.output_number))
        self.variances = np.empty((0, self.output_number))
        self.value_positions = np.empty((0, self.output_number, self.oput_dim))
        self.input_space_bounds = np.array(input_space_bounds)
        #self.hyperparameters = None
        self.gp_initialized = False
        self.cost_function_parameters = None
        self.cost_function = None
        self.consider_costs = False
        GPOptimizer.__init__(self,
                input_space_dimension,
                input_space_bounds
                )

    def get_data_fvGP(self):
        """
        Provides a way to access the current class variables.

        Returns
        -------
        dict
            Dictionary containing the input dim, output dim, output number,
            x & y data, measurement variances, measurement value positions,
            hyperparameters, cost function parameters and consider costs
            class attributes. Note that if tell() has not been called, many
            of these returned values will be None.
        """

        res = self.get_data
        res["output number"] = self.output_number
        res["output dim"] = self.oput_dim
        res["measurement value positions"] =  self.value_positions
        return res


    def tell(self, x, y, variances=None, value_positions=None):
        """
        This function can tell() the gp_optimizer class
        the data that was collected. The data will instantly be use to update the gp_data
        if a GP was previously initialized

        Parameters:
        -----------
            x (2d numpy array):                A 2d array of all the points in the data
            values (2d numpy array):           A 2d array of all the values measured at the associated points

        Optional Parameters:
        --------------------
            variances (2d numpy array):         A 2d array of all the variances of the measured points
            value_positions (3d numpy array):   A 3d numpy array that stores a 
                                                2d array of locations for each each data point
                                                    e.g. 
                                                    * 2 data points with 2 ouputs in 1d:
                                                      value_posiitons = np.array([
                                                      [[0],[1]]
                                                      [[0],[1]]
                                                      ])
                                                    * 2 data points with 3 ouputs in 2d:
                                                      value_posiitons = np.array([
                                                      [[0,1],[2,3],[4,5]]
                                                      [[0,2],[4,2],[7,8]]
                                                      ])

        Returns:
        --------
            no returns
        """
        self.points = x
        self.values = y
        self.variances = variances
        self.value_positions = value_positions

        if self.gp_initialized is True: self.update_fvgp()

##############################################################
    def init_fvgp(
        self,
        init_hyperparameters,
        compute_device="cpu",
        gp_kernel_function=None,
        gp_mean_function=None,
        sparse=False
    ):
        """
        Function to initialize the GP if it has not already been initialized
        Parameters:
        -----------
            init_hyperparameters:   1d numpy array containing the initial guesses for the hyperparemeters
        Optional Parameters:
        --------------------
            compute_device, default = cpu, others = "gpu"
            gp_kernel_function, default = fvGP default
            gp_mean_function, default = fvGP default, i.e. average of data
            sparse, default = False
        """
        if self.gp_initialized is False:
            fvGP.__init__(
            self,
            self.iput_dim,
            self.oput_dim,
            self.output_number,
            self.points,
            self.values,
            init_hyperparameters,
            value_positions = self.value_positions,
            variances = self.variances,
            compute_device = compute_device,
            gp_kernel_function = gp_kernel_function,
            gp_mean_function = gp_mean_function,
            sparse = sparse,
            )
            self.gp_initialized = True
        else: print("fvGP already initialized")

##############################################################
    def update_fvgp(self):
        """
        This function updates the data in the fvGP, tell(...) will call this function automatically if
        GP is already intialized
        Paramaters:
        -----------
            no input parameters
        """
        self.update_fvgp_data(
            self.points,
            self.values,
            value_positions = self.value_positions,
            variances = self.variances)




#!/usr/bin/env python

"""Tests for `gpcam` package."""
import numpy as np
from gpcam.gp_optimizer import GPOptimizer
import matplotlib.pyplot as plt
import unittest

def ac_func1(x,obj):
    r1 = obj.posterior_mean(x)["f(x)"]
    r2 = obj.posterior_covariance(x)["v(x)"]
    m_index = np.argmin(obj.data_y)
    m = obj.data_x[m_index]
    std_model = np.sqrt(r2)
    return -(r1 + 3.0 * std_model)

class TestgpCAM(unittest.TestCase):
    """Tests for `gpcam` package."""

    def setUp(self,dim = 2):
        """Set up test fixtures, if any."""
        x = np.random.rand(100,dim)
        y = np.sin(x[:,0])
        ######################################################
        ######################################################
        ######################################################
        index_set_bounds = np.array([[0.,1.],[0.,1.]])
        hyperparameter_bounds = np.array([[0.001,1e9],[0.001,100],[0.001,100]])
        hps_guess = np.ones((3))
        ###################################################################################
        gp = GPOptimizer(dim,index_set_bounds)
        gp.tell(x,y)
        gp.init_gp(hps_guess)
        gp.train_gp(hyperparameter_bounds)

    def test_single_task(self,dim = 2, N = 100):
        """Test something."""
        x = np.random.rand(100,dim)
        y = np.sin(x[:,0])
        ######################################################
        ######################################################
        ######################################################
        index_set_bounds = np.array([[0.,1.],[0.,1.]])
        hyperparameter_bounds = np.array([[0.001,1e9],[0.001,100],[0.001,100]])
        hps_guess = np.ones((3))
        ###################################################################################
        gp = GPOptimizer(dim,index_set_bounds)
        gp.tell(x,y)
        gp.init_gp(hps_guess)
        gp.train_gp(hyperparameter_bounds)
        ######################################################
        ######################################################
        ######################################################
        print("evaluating acquisition function at [0.5,0.5,0.5]")
        print("=======================")
        r1 = gp.evaluate_acquisition_function(np.array([0.5,0.5]),acquisition_function = "shannon_ig")
        r2 = gp.evaluate_acquisition_function(np.array([0.5,0.5]),acquisition_function = ac_func1)
        print("results: ",r1,r2)
        input("Continue with ENTER")
        print("getting data from gp optimizer:")
        print("=======================")
        r = gp.get_data()
        print(r)
        input("Continue with ENTER")
        print("ask()ing for new suggestions")
        print("=======================")
        r = gp.ask()
        print(r)
        input("Continue with ENTER")
        print("getting the maximum (remember that this means getting the minimum of -f(x)):")
        print("=======================")
        r = gp.ask(acquisition_function = "maximum")
        print(r)
        print("getting the minimum:")
        print("=======================")
        r = gp.ask(acquisition_function = "minimum")
        print(r)
        input("Continue with ENTER")
        print("Writing interpolation to file...")
        print("=======================")


        ar3d = np.empty((50,50))
        l = np.empty((50*50,4))
        x = np.linspace(0,1,50)
        y = np.linspace(0,1,50)
        counter = 0
        for i in range(50):
            print("done ",((i+1.0)/50.0)*100.," percent")
            for j in range(50):
                res = gp.posterior_mean(np.array([[x[i],y[j]]]))
                ar3d[i,j] = res["f(x)"]
                l[counter,0] = x[i]
                l[counter,1] = y[j]
                l[counter,3] = res["f(x)"] / 10000.0
                counter += 1

        file_name = "data_list.csv"
        np.savetxt(file_name,l, delimiter = ",",header = 'x coord, y coord, z_coord, scalar')
        print("==================================================")
        print("data cube written in 'data_list.csv'; you can use paraview to visualize it")
        print("END")




import time
from time import sleep
import random
import numpy as np
from random import seed
from random import randint
import itertools
import math
import uuid
import time
import datetime

class gpData:
    """
    Data Class

    dataset is a list of dictionaries
    data arrays are numpy arrays
    """
    def __init__(self, dim, parameter_bounds, output_number = None, output_dim = None):
        self.dim = dim
        self.parameter_bounds = parameter_bounds
        self.dataset = []
        self.output_number = output_number
        self.output_dim =output_dim

    ###############################################################
    #either create random data or communicate a data set (use translate2data for numpy arrays)
    def create_random_dataset(self, length):
        """
        creates random data of "length" and creates a dataset
        """
        self.x = self._create_random_points(length)
        self.point_number = len(self.x)
        self.dataset = self.inject_arrays(self.x)

    def inject_dataset(self,dataset):
        """
        initializes a previously-collected dataset
        !!!for intitiation only!!! just use the "+" operator to update the existing dataset
        """
        self.point_number = len(self.dataset)
        self.dataset = dataset

    def inject_arrays(self, x, info = None):
        """
        translates numpy arrays to the data format
        """
        data = []
        for i in range(len(x)):
            data.append(self.npy2dataset_entry(x[i]))
            if info is not None: data[i].update(info[i])
        return data
    ###############################################################
    def npy2dataset_entry(self, x):
        """
        parameters:
        -----------
            x ... 1d numpy array
        """
        d = {}
        d["position"] = x
        d["value"] = None
        d["variance"] = None
        d["cost"] = None
        d["id"] = str(uuid.uuid4())
        d["time stamp"] = time.time()
        d["date time"] = datetime.datetime.now().strftime("%d/%m/%Y_%H:%M%S")
        d["measured"] = False
        d["posterior variance"] = None #post_var
        d["hyperparameters"] = None #hps
        return d
    ###############################################################
    ###Printing####################################################
    ###############################################################
    def print_data(self, data):
        np.set_printoptions(precision=5)
        for idx in range(len(data)):
            print(idx, " ==> ")
            for key in list(data[idx]):
                if isinstance(data[idx][key], dict):
                    print("     ", key, " :")
                    for key2 in data[idx][key]:
                        print("          ", key2, " : ", data[idx][key][key2])
                else:
                    print("     ", key, " : ", data[idx][key])
    ################################################################
    ########Extracting##############################################
    ################################################################
    def extract_data(self):
        x = self.extract_points_from_data()
        y = self.extract_values_from_data()
        v = self.extract_variances_from_data()
        t = self.extract_times_from_data()
        c = self.extract_costs_from_data()
        return x,y,v,t,c

    def extract_points_from_data(self):
        self.point_number = len(self.dataset)
        P = np.zeros((self.point_number, self.dim))
        for idx_data in range(self.point_number):
            P[idx_data] = self.dataset[idx_data]["position"]
        return P

    def extract_values_from_data(self):
        self.point_number = len(self.dataset)
        M= np.zeros((self.point_number))
        for idx_data in range(self.point_number):
            M[idx_data] = self.dataset[idx_data]["value"]
        return M

    def extract_variances_from_data(self):
        self.point_number = len(self.dataset)
        Variance = np.zeros((self.point_number))
        for idx_data in range(self.point_number):
            if self.dataset[idx_data]["variance"] is None: return None
            Variance[idx_data] = self.dataset[idx_data]["variance"]
        return Variance

    def extract_costs_from_data(self):
        self.point_number = len(self.dataset)
        Costs = []
        for idx in range(self.point_number):
            Costs.append(self.dataset[idx]["cost"])
        return Costs

    def extract_times_from_data(self):
        self.point_number = len(self.dataset)
        times = np.zeros((self.point_number))
        for idx_data in range(self.point_number):
            times[idx_data] = self.dataset[idx_data]["time stamp"]
        return times

    ###############################################################
    #######Creating################################################
    ###############################################################
    def _create_random_x(self):
        return np.random.uniform(low = self.parameter_bounds[:,0],
                                 high =self.parameter_bounds[:,1],
                                 size = self.dim)
    def _create_random_points(self, length):
        x = np.empty((length,self.dim))
        for i in range(length):
            x[i,:] = self._create_random_x()
        return x
    ###############################################################
    #########Cleaning##############################################
    ###############################################################
    def check_incoming_data(self):
        for entry in self.dataset:
            if entry["value"] is None:
                raise Exception("Entry with no secified value entry in communicated list of data dictionaries")

    def clean_data_NaN(self):
        for entry in self.dataset:
            if self._nan_in_dict(entry):
                print("CAUTION, NaN detected in data")
                self.dataset.remove(entry)
        self.point_number = len(self.data_set)

    def nan_in_dataset(self):
        for entry in self.dataset:
            if self._nan_in_dict(entry):
                return True
        return False


    def _nan_in_dict(self,dictionary):
        is_nan = False
        try:
            for key in dictionary:
                if type(dictionary[key]) is dict:
                    is_nan = self._nan_in_dict(dictionary[key])
                elif type(dictionary[key]) is float and math.isnan(dictionary[key]):
                    is_nan = True
                elif type(dictionary[key]) is np.ndarray and any(np.isnan(dictionary[key])):
                    is_nan = True
                elif type(dictionary[key]) is list and any(np.isnan(np.asarray(dictionary[key]))):
                    is_nan = True
                else:
                    is_nan = False
        except:
            pass
        return is_nan
######################################################################
######################################################################
######################################################################
######################################################################
######################################################################
######################################################################
######################################################################
######################################################################
######################################################################
class fvgpData(gpData):
    def create_random_dataset(self, length):
        if self.output_number is None or self.output_dim is None:
            raise Exception("When initializing the data class for a multi-output GP, \
                    please provide output_number AND an output_dim parameters.")

        self.x = self._create_random_points(length)
        self.point_number = len(self.x)
        self.dataset = self.inject_arrays(self.x)

    def inject_dataset(self,dataset):
        """
        initializes a previously-collected dataset
        !!!for intitiation only!!! just use the "+" operator to update the existing dataset
        """
        if self.output_number is None or self.output_dim is None:
            raise Exception("When initializing the data class for a multi-output GP, \
                    please provide output_number AND an output_dim parameters.")

        self.point_number = len(self.dataset)
        self.dataset = dataset

    def inject_arrays(self, x, info = None):
        """
        translates numpy arrays to the data format
        """
        data = []
        for i in range(len(x)):
            data.append(self.npy2dataset_entry(x[i]))
            if info is not None: data[i].update(info[i])
        return data

    def npy2dataset_entry(self, x):
        """
        parameters:
        -----------
        x ... 1d numpy array
        """
        d = {}
        d["position"] = x
        d["values"] = None
        d["variances"] = None
        d["value positions"] = None
        d["cost"] = None
        d["id"] = str(uuid.uuid4())
        d["time stamp"] = time.time()
        d["date time"] = datetime.datetime.now().strftime("%d/%m/%Y_%H:%M%S")
        d["measured"] = False
        d["posterior variances"] = None
        d["hyperparameters"] = None
        return d

    ################################################################
    ################################################################
    ################################################################
    def extract_data(self):
        x = self.extract_points_from_data()
        y = self.extract_values_from_data()
        v = self.extract_variances_from_data()
        t = self.extract_times_from_data()
        c = self.extract_costs_from_data()
        vp= self.extract_value_positions_from_data()
        return x,y,v,t,c,vp


    def extract_value_positions_from_data(self):
        self.point_number = len(self.dataset)
        VP = np.zeros((self.point_number, self.output_number, self.output_dim))
        for idx_data in range(self.point_number):
            if ("value positions" in self.dataset[idx_data]):
                VP[idx_data] = self.dataset[idx_data]["value positions"]
            else:
                VP[idx_data] = self.dataset[idx_data - 1]["value positions"]
        return VP

    def extract_values_from_data(self):
        self.point_number = len(self.dataset)
        M= np.zeros((self.point_number,self.output_number))
        for idx_data in range(self.point_number):
            M[idx_data] = self.dataset[idx_data]["values"]
        return M

    def extract_variances_from_data(self):
        self.point_number = len(self.dataset)
        Variance = np.zeros((self.point_number,self.output_number))
        for idx_data in range(self.point_number):
            if self.dataset[idx_data]["variances"] is None: return None
            Variance[idx_data] = self.dataset[idx_data]["variances"]
        return Variance





#/usr/bin/env python
import time
from time import strftime
import numpy as np
from gpcam.data import gpData
from gpcam.data import fvgpData
from gpcam.gp_optimizer import GPOptimizer
from gpcam.gp_optimizer import fvGPOptimizer

#todo: autonomous experimenter for fvgp, data class for fvgp


class AutonomousExperimenterGP():
    """
    class AutonomousExperimenterGP:
    executes the autonomous loop for a single-task GP
    use class AutonomousExperimenterfvGP for multi-task experiments
    Parameters:
    -----------
        * parameter_bounds
        * instrument_func
        * hyperparameters
        * hyperparameter_bounds
    Optional Parameters:
    --------------------
        * init_dataset_size = None: int or None, None means you have to provide intial data
        * acq_func = "covariance": acquisition function to be maximized in search of new measurements
        * cost_func = None
        * cost_update_func = None
        * cost_func_params = {}
        * kernel_func = None
        * prior_mean_func = None
        * run_every_iteration = None
        * x = None, y = None, v = None: inital data can be supplied here
        * append_data_after_send = False: Append data or communiate entire dataset
        * compute_device = "cpu"
        * sparse = False
        * training_dask_client = None
        * acq_func_opt_dask_client = None
    """
    def __init__(self,
            parameter_bounds,
            instrument_func,
            hyperparameters,
            hyperparameter_bounds,
            init_dataset_size = None,
            acq_func = "covariance",
            cost_func = None,
            cost_update_func = None,
            cost_func_params = {},
            kernel_func = None,
            prior_mean_func = None,
            run_every_iteration = None,
            x = None, y = None, v = None, dataset = None,
            append_data_after_send = False,
            compute_device = "cpu",
            sparse = False,
            training_dask_client = None,
            acq_func_opt_dask_client = None
            ):
        dim = len(parameter_bounds)
        self.instrument_func = instrument_func
        self.hyperparameter_bounds = hyperparameter_bounds
        self.acq_func = acq_func
        self.cost_func = cost_func
        self.cost_update_func = cost_update_func
        self.kernel_func = kernel_func
        self.prior_mean_func = prior_mean_func
        self.run_every_iteration = run_every_iteration
        self.append = append_data_after_send
        self.async_train = False
        self.training_dask_client = training_dask_client
        self.acq_func_opt_dask_client = acq_func_opt_dask_client
        ################################
        #getting the data ready#########
        ################################
        if init_dataset_size is None and x is None and dataset is None:
            raise Exception("Either provide length of initial data or an inital dataset")
        self.data = gpData(dim, parameter_bounds)
        if (x is None or y is None) and dataset is None:
            self.data.create_random_dataset(init_dataset_size)
            self.data.dataset = self.instrument_func(self.data.dataset)
        elif (x is None or y is None) and dataset is not None:
            self.data.inject_dataset(list(np.load(dataset, allow_pickle = True)))
            hyperparameters = self.data.dataset[-1]["hyperparameters"]
        elif x is not None and y is not None:
            self.data.inject_arrays(x)
            self.data.dataset = self.instrument_func(self.data.dataset)
        else: raise Exception("No viable option for data given!")
        if self.data.nan_in_dataset(): self.data.clean_data_NaN()
        self.x, self.y, self.v, self.t, self.c = self.data.extract_data()
        self.init_dataset_size = len(self.x)
        ######################
        ######################
        ######################
        self.gp_optimizer = GPOptimizer(dim,parameter_bounds)
        self.gp_optimizer.tell(self.x, self.y,variances = self.v)
        self.gp_optimizer.init_gp(hyperparameters, compute_device = compute_device,
            gp_kernel_function = self.kernel_func,
            gp_mean_function = self.prior_mean_func,
            sparse = sparse)
        #init costs
        self._init_costs(cost_func_params)
        print("##################################################################################")
        print("Initialization successfully concluded")
        print("now train(...) or train_async(...), and then go(...)")
        print("##################################################################################")
    ###################################################################################
    def train(self,pop_size = 10, tol = 1e-6, max_iter = 20, method = "global"):
        self.gp_optimizer.train_gp(
        self.hyperparameter_bounds,
        method = method, pop_size = pop_size,
        tolerance = tol, max_iter = max_iter)

    def train_async(self,pop_size = 10,tol = 1e-6, max_iter = 20, dask_client = None):
        self.opt_obj = self.gp_optimizer.train_gp_async(
        self.hyperparameter_bounds,pop_size = pop_size,
        tolerance = tol,max_iter = max_iter,
        dask_client = dask_client
        )
        self.async_train = True

    def kill_training(self):
        print("async training is being killed")
        if self.async_train: self.gp_optimizer.stop_async_train(self.opt_obj)
        self.async_train = False

    def kill_client(self):
        if self.async_train: self.gp_optimizer.kill_async_train(self.opt_obj)

    def update_hps(self):
        if self.async_train: self.gp_optimizer.update_hyperparameters(self.opt_obj)
        print("The Autonomus Experimenter updated the Hyperparameters")
        print("hps: ", self.gp_optimizer.hyperparameters)

    def _init_costs(self,cost_func_params):
        self.gp_optimizer.init_cost(self.cost_func,cost_func_params,
                cost_update_function = self.cost_update_func)

    def tell(self, x,y,v, vp  = None):
        if vp is None: self.gp_optimizer.tell(x, y,variances = v)
        else: self.gp_optimizer.tell(x, y,variances = v, value_positions = vp)

    def extract_data(self):
        x,y,v,t,c = self.data.extract_data()
        return x,y,v,t,c, None
    ###################################################################################
    def go(self, N = 1e15, breaking_error = 1e-50,
            retrain_globally_at = [100,400,1000],
            retrain_locally_at = [20,40,60,80,100,200,400,1000],
            retrain_async_at = [1000,2000,5000,10000],
            retrain_callable_at = [],
            update_cost_func_at = [],
            acq_func_opt_setting = lambda number: "global" if number % 2 == 0 else "local",
            training_opt_callable = None,
            training_opt_max_iter = 20,
            training_opt_pop_size = 10,
            training_opt_tol      = 1e-6,
            acq_func_opt_max_iter = 20,
            acq_func_opt_pop_size = 20,
            acq_func_opt_tol      = 1e-6,
            acq_func_opt_tol_adjust = [True,0.1],
            number_of_suggested_measurements = 1,
            ):
        """
        function to start the autonomous-data-acquisition loop
        optional parameters:
        -----------
            * N = 1e15 ... run N iterations
            * breaking_error = 1e-15 ... run until breaking_error is achieved
            * retrain_globally_at = [100,400,1000]
            * retrain_locally_at = [20,40,60,80,100,200,400,1000]
            * retrain_async_at = [1000,2000,5000,10000]
            * retrain_callable = []: if this is not an empty list, "training_opt_callable" has to be provided
            * update_cost_func_at = []: list containing numbers when the cost function is updated
            * search_setting = lambda function to decide when global, local, hgdl, other ask()
            * training_opt_callable = None, callable
            * training_opt_max_iter = 20
            * training_opt_pop_size = 10
            * training_opt_tol      = 1e-6
            * acq_func_opt_max_iter = 20
            * acq_func_opt_pop_size = 20
            * acq_func_opt_tol      = 1e-6
            * acq_func_opt_tol_adjust = [True, 0.1]
            * number_of_suggested_measurements = 1
            * compute_device = cpu
            * sparse = False
            * training_dask_client = None
            * acq_func_opt_dask_client = None
        """
        start_time = time.time()
        start_date_time = strftime("%Y-%m-%d_%H_%M_%S", time.localtime())
        print("Date and time:       ", start_date_time)

        for i in range(self.init_dataset_size,int(N)):
            n_measurements = len(self.x)
            print("")
            print("")
            print("")
            print("====================")
            print("====================")
            print("iteration: ",i)
            print("Run Time: ", time.time() - start_time, "     seconds")
            print("Number of measurements: ", n_measurements)
            print("====================")
            #ask() for new suggestions
            current_position = self.x[-1]
            print("hps: ",self.gp_optimizer.hyperparameters)
            res = self.gp_optimizer.ask(
                    position = current_position,
                    n = number_of_suggested_measurements,
                    acquisition_function = self.acq_func,
                    cost_function = self.cost_func,
                    bounds = None,
                    method = acq_func_opt_setting(i),
                    pop_size = acq_func_opt_pop_size,
                    max_iter = acq_func_opt_max_iter,
                    tol = acq_func_opt_tol,
                    dask_client = self.acq_func_opt_dask_client)
            #########################
            next_measurement_points = res["x"]
            func_evals = res["f(x)"]
            post_var = self.gp_optimizer.posterior_covariance(next_measurement_points)["v(x)"]
            error = np.max(np.sqrt(post_var[0]))
            if acq_func_opt_tol_adjust[0]:
                acq_func_opt_tol = abs(func_evals[0]) * acq_func_opt_tol_adjust[1]
                print("variance optimization tolerance changed to: ", acq_func_opt_tol)
            print("Next points to be requested: ")
            print(next_measurement_points)
            #update and tell() new data
            info  = [{"hyperparameters" : self.gp_optimizer.hyperparameters,
                      "posterior variance" : post_var[i]} for i in range(len(next_measurement_points))]
            new_data = self.data.inject_arrays(next_measurement_points, info = info)
            print("Sending request to instrument ...")
            if self.append: self.data.dataset = self.data.dataset + self.instrument_func(new_data)
            else: self.data.dataset = self.instrument_func(self.data.dataset + new_data)
            print("Data received")
            print("Checking if data is clean ...")
            if self.data.nan_in_dataset(): self.data.clean_data_NaN()
            self.data.check_incoming_data()
            print("done")
            #update arrays and the gp_optimizer
            self.x,self.y, self.v, self.t, self.c,vp = self.extract_data()
            print("Communicating new data to GP")
            self.tell(self.x, self.y, self.v, vp)
            ###########################
            #train()
            print("Training ...")
            if len(self.x) in retrain_async_at:
                self.kill_training()
                self.opt_obj = self.train_async(pop_size = 10,tol = 1e-6,max_iter = 20,
                                 dask_client = self.training_dask_client)
            elif len(self.x) in retrain_globally_at:
                self.kill_training()
                print("Fresh optimization from scratch via global optimization")
                self.train(pop_size = 10,tol = 1e-6,max_iter = 20, method = "global")
            elif len(self.x) in retrain_locally_at:
                self.kill_training()
                print("Fresh optimization from scratch via global optimization")
                self.train(pop_size = 10,tol = 1e-6, max_iter = 20, method = "local")
            elif len(self.x) in retrain_callable_at:
                self.kill_training()
                print("Fresh optimization from scratch via global optimization")
                self.train(pop_size = 10,tol = 1e-6,max_iter = 20, method = training_opt)
            else:
                self.update_hps()
                print("No training in this round but I tried to update the hyperparameters")
            ###save some data
            try: np.save('Data_'+ start_date_time, self.data.dataset)
            except Exception as e: print("Data not saved due to ", str(e))
            ###########################
            #cost update
            if i in update_cost_func_at: self.gp_optimizer.update_cost_function(self.c)

            if error < breaking_error: break
        self.kill_client()
        print("====================================================")
        print("The autonomous experiment was concluded successfully")
        print("====================================================")

###################################################################################
###################################################################################
###################################################################################
###################################################################################
###################################################################################
###################################################################################
class AutonomousExperimenterFvGP(AutonomousExperimenterGP):
    def __init__(self,
            parameter_bounds,
            output_number,
            output_dim,
            instrument_func,
            hyperparameters,
            hyperparameter_bounds,
            init_dataset_size = None,
            acq_func = "covariance",
            cost_func = None,
            cost_update_func = None,
            cost_func_params = {},
            kernel_func = None,
            prior_mean_func = None,
            run_every_iteration = None,
            x = None, y = None, v = None, vp = None, dataset = None,
            append_data_after_send = False,
            compute_device = "cpu",
            sparse = False,
            training_dask_client = None,
            acq_func_opt_dask_client = None
            ):
        dim = len(parameter_bounds)
        self.instrument_func = instrument_func
        self.hyperparameters = hyperparameters
        self.hyperparameter_bounds = hyperparameter_bounds
        self.acq_func = acq_func
        self.cost_func = cost_func
        self.cost_update_func = cost_update_func
        self.kernel_func = kernel_func
        self.prior_mean_func = prior_mean_func
        self.run_every_iteration = run_every_iteration
        self.append = append_data_after_send
        self.async_train = False
        self.training_dask_client = training_dask_client
        self.acq_func_opt_dask_client = acq_func_opt_dask_client
        ################################
        #getting the data ready#########
        ################################
        if init_dataset_size is None and x is None and dataset is None:
            raise Exception("Either provide length of initial data or an inital dataset")
        self.data = fvgpData(dim, parameter_bounds,
                output_number = output_number, output_dim = output_dim)
        if (x is None or y is None) and dataset is None:
            self.data.create_random_dataset(init_dataset_size)
            self.data.dataset = self.instrument_func(self.data.dataset)
        elif (x is None or y is None) and dataset is not None:
            self.data.inject_dataset(list(np.load(dataset, allow_pickle = True)))
            self.hyperparameters = self.data.dataset[-1]["hyperparameters"]
        elif x is not None and y is not None:
            self.data.inject_arrays(x)
            self.data.dataset = self.instrument_func(self.data.dataset)
        else: raise Exception("No viable option for data given!")
        if self.data.nan_in_dataset(): self.data.clean_data_NaN()
        self.x, self.y, self.v, self.t, self.c, self.vp = self.data.extract_data()
        self.init_dataset_size = len(self.x)
        ######################
        ######################
        ######################
        self.gp_optimizer = fvGPOptimizer(dim,output_dim,output_number,parameter_bounds)
        self.gp_optimizer.tell(self.x, self.y,variances = self.v,value_positions = self.vp)
        self.gp_optimizer.init_fvgp(self.hyperparameters,compute_device = compute_device,
            gp_kernel_function = self.kernel_func,
            gp_mean_function = self.prior_mean_func,
            sparse = sparse)
        #init costs
        self._init_costs(cost_func_params)
        print("##################################################################################")
        print("Initialization successfully concluded")
        print("now train(...) or train_async(...), and then go(...)")
        print("##################################################################################")

    def extract_data(self):
        x,y,v,t,c,vp = self.data.extract_data()
        return x,y,v,t,c,vp


#class AutonomousExperimenterEnsembleGP():



#/usr/bin/env python
import numpy as np
from gpcam.autonomous_experimenter import AutonomousExperimenterGP

def instrument(data):
    for entry in data:
        entry["value"] = np.sin(np.linalg.norm(entry["position"]))
    return data

my_ae = AutonomousExperimenterGP(np.array([[0,10],[0,10]]),instrument,np.ones((3)),np.array([[0.001,100],[0.001,100],[0.001,100]]),init_dataset_size= 10)
my_ae.train()
input("training done")
my_ae.go(N = 200)




#!/usr/bin/env python

"""The setup script."""

from setuptools import setup, find_packages
from os import path
import sys

with open('README.rst') as readme_file:
    readme = readme_file.read()

with open('HISTORY.rst') as history_file:
    history = history_file.read()

#requirements = ["requirements.txt"]

#setup_requirements = ["requirements.txt"]

#test_requirements = ["requirements.txt"]


here = path.abspath(path.dirname(__file__))
with open(path.join(here, 'requirements.txt')) as requirements_file:
    # Parse requirements.txt, ignoring any commented-out lines.
    requirements = [line for line in requirements_file.read().splitlines()
                    if not line.startswith('#')]


setup(
    author="Marcus Michael Noack",
    author_email='MarcusNoack@lbl.gov',
    python_requires='>=3.7',
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Intended Audience :: Developers',
        ####License here:
        ####e.g. 'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
        'Natural Language :: English',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
    ],
    description="gpCAM is a code for autonomous data acquisition",
    install_requires=requirements,
    #license="GNU General Public License v3",
    long_description=readme + '\n\n' + history,
    include_package_data=True,
    keywords='gpcam',
    name='gpcam',
    packages=find_packages(include=['gpcam', 'gpcam.*']),
    test_suite='tests',
    url='https://github.com/Marcus Noack/gpcam',
    version='7.0.5',
    zip_safe=False,
)
